#include "MappedGridOperators.h"

#include "MappedGridOperatorsInclude.h"

void MappedGridOperators::
applyBCtangentialComponent(realMappedGridFunction & u, 
			   const int side,
			   const int axis,
			   const Index & Components,
			   const BCTypes::BCNames & bcType,
			   const int & bc,
			   const real & scalarData,
			   const RealArray & arrayData,
                           const RealArray & arrayDataD,
			   const realMappedGridFunction & gfData,
			   const real & t,
                           const IntegerArray & uC, const IntegerArray & fC, const IntegerDistributedArray & mask,
			   const BoundaryConditionParameters & bcParameters,
			   const BoundaryConditionOption bcOption,
			   const int & grid  )
{
  if( !boundaryNormalsUsed && bcType==BCTypes::tangentialComponent )
  {
    boundaryNormalsUsed=TRUE;
    mappedGrid.update(MappedGrid::THEvertexBoundaryNormal);
  }
  if( !boundaryTangentsUsed && 
     ( bcType==BCTypes::tangentialComponent0 || bcType==BCTypes::tangentialComponent1) )
  {
    boundaryTangentsUsed=TRUE;
    mappedGrid.update(MappedGrid::THEcenterBoundaryTangent);
  }
  MappedGrid & c = mappedGrid;
  RealDistributedArray & uA = u;
  RealDistributedArray & normal  = mappedGrid.vertexBoundaryNormal(side,axis);   // make centerBoundaryNormal ***
  RealDistributedArray & tangent = mappedGrid.centerBoundaryTangent(side,axis);
  int n1,n2,n3,m1,m2,m3,nt,ndt;

  Index I1,I2,I3;
  getGhostIndex( c.indexRange(),side,axis,I1,I2,I3,bcParameters.lineToAssign,bcParameters.extraInTangentialDirections);

  RealDistributedArray uDotN(I1,I2,I3);

  switch (bcType)
  {
  case tangentialComponent0:
  case tangentialComponent1:


    // tangent vectors are stored as : tangent(I1,I2,I3,0:nd-1,0:nd-2)  (last index = tangent 0 or 1)
    nt= bcType==BCTypes::tangentialComponent0 ? 0 : 1;
    ndt=nt*numberOfDimensions;  // for indexing into the tangent array (last two components are merged)
    if( numberOfDimensions==1 )
    {
      cout << "applyBoundaryCondition::ERROR: cannot apply tangentialComponent" << nt << " BC in 1D\n";
      {throw "error";}
    }
    //
    // to set the tangential component to g:
    //       u <- u + (g-(n.u)) n
    //
    getVelocityComponents( n1,n2,n3,m1,m2,m3, u,bcParameters,"tangentialComponent[0,1]",uC,fC  );
	
    if( numberOfDimensions==2 )
    {
      uDotN(I1,I2,I3)=uA(I1,I2,I3,n1)*tangent(I1,I2,I3,0+ndt)+uA(I1,I2,I3,n2)*tangent(I1,I2,I3,1+ndt);
    }
    else
    {
      uDotN(I1,I2,I3)=  uA(I1,I2,I3,n1)*tangent(I1,I2,I3,0+ndt)
	+uA(I1,I2,I3,n2)*tangent(I1,I2,I3,1+ndt) 
	+uA(I1,I2,I3,n3)*tangent(I1,I2,I3,2+ndt); 
    }

    if( twilightZoneFlow ) 
    { // In this case we want to specify the value for n.u
      if( numberOfDimensions==2 )
	uDotN(I1,I2,I3)-=(*e)(c,I1,I2,I3,m1,t)*tangent(I1,I2,I3,0+ndt)
	  +(*e)(c,I1,I2,I3,m2,t)*tangent(I1,I2,I3,1+ndt);
      else 
	uDotN(I1,I2,I3)-=(*e)(c,I1,I2,I3,m1,t)*tangent(I1,I2,I3,0+ndt)
	  +(*e)(c,I1,I2,I3,m2,t)*tangent(I1,I2,I3,1+ndt) 
	  +(*e)(c,I1,I2,I3,m3,t)*tangent(I1,I2,I3,2+ndt);
    }
    else if( bcOption==scalarForcing )
    {
      if( scalarData != 0. )
	uDotN(I1,I2,I3)-=scalarData;   // use user supplied constant value
    }
    else if( bcOption==arrayForcing )
    {
      #ifdef USE_PPP
	Overture::abort("MappedGridOperators::applyBCtangential:ERROR finish me Bill!");
      #else 
      if( arrayDataD.getBase(0)<=I1.getBase() && arrayDataD.getBound(0)>=I1.getBound() &&
	  arrayDataD.getBase(1)<=I2.getBase() && arrayDataD.getBound(1)>=I2.getBound() &&
	  arrayDataD.getBase(2)<=I3.getBase() && arrayDataD.getBound(2)>=I3.getBound() &&
	  arrayDataD.getBase(3)<=min(m1,m2,m3) && arrayDataD.getBound(3)>=max(m1,m2,m3) )
      {
	if( numberOfDimensions==2 ) 
	  uDotN(I1,I2,I3)-=(arrayDataD(I1,I2,I3,m1)*tangent(I1,I2,I3,0+ndt)+
			    arrayDataD(I1,I2,I3,m2)*tangent(I1,I2,I3,1+ndt));
	else if( numberOfDimensions==3 )
	  uDotN(I1,I2,I3)-=(arrayDataD(I1,I2,I3,m1)*tangent(I1,I2,I3,0+ndt)+
			    arrayDataD(I1,I2,I3,m2)*tangent(I1,I2,I3,1+ndt)+
			    arrayDataD(I1,I2,I3,m3)*tangent(I1,I2,I3,2+ndt));
      }
      else if( side<=arrayData.getBound(1) && axis<=arrayData.getBound(2) && grid<=arrayData.getBound(3) )
      {
	if( numberOfDimensions==2 ) 
	  uDotN(I1,I2,I3)-=arrayData(m1,side,axis,grid)*tangent(I1,I2,I3,0+ndt)
	    +arrayData(m2,side,axis,grid)*tangent(I1,I2,I3,1+ndt);
	else 
	  uDotN(I1,I2,I3)-=arrayData(m1,side,axis,grid)*tangent(I1,I2,I3,0+ndt)
	    +arrayData(m2,side,axis,grid)*tangent(I1,I2,I3,1+ndt) 
	    +arrayData(m3,side,axis,grid)*tangent(I1,I2,I3,2+ndt);
      }
      else
      {
	if( numberOfDimensions==2 ) 
	  uDotN(I1,I2,I3)-=arrayData(m1)*tangent(I1,I2,I3,0+ndt)
	    +arrayData(m2)*tangent(I1,I2,I3,1+ndt);
	else 
	  uDotN(I1,I2,I3)-=arrayData(m1)*tangent(I1,I2,I3,0+ndt)
	    +arrayData(m2)*tangent(I1,I2,I3,1+ndt) 
	    +arrayData(m3)*tangent(I1,I2,I3,2+ndt);
      }
      #endif
    }
    else if( bcOption==gridFunctionForcing )
    {  // use user supplied variable values
      if( gfData.getComponentDimension(0) < numberOfDimensions )   
	uDotN(I1,I2,I3)-=gfData(I1,I2,I3,m1);
      else if( numberOfDimensions==2 ) 
	uDotN(I1,I2,I3)-=gfData(I1,I2,I3,m1)*tangent(I1,I2,I3,0+ndt)
	  +gfData(I1,I2,I3,m2)*tangent(I1,I2,I3,1+ndt);
      else 
	uDotN(I1,I2,I3)-=gfData(I1,I2,I3,m1)*tangent(I1,I2,I3,0+ndt)
	  +gfData(I1,I2,I3,m2)*tangent(I1,I2,I3,1+ndt) 
	  +gfData(I1,I2,I3,m3)*tangent(I1,I2,I3,2+ndt);
    }
    else
      {throw "Invalid value for bcOption! (tangentialComponent)";}

    WHERE_MASK( uA(I1,I2,I3,n1)-=uDotN(I1,I2,I3)*tangent(I1,I2,I3,0+ndt); )
    WHERE_MASK( uA(I1,I2,I3,n2)-=uDotN(I1,I2,I3)*tangent(I1,I2,I3,1+ndt); )
    if( numberOfDimensions==3 )
      WHERE_MASK( uA(I1,I2,I3,n3)-=uDotN(I1,I2,I3)*tangent(I1,I2,I3,2+ndt); )

    break;
	
  case tangentialComponent:
    //
    // to set the tangential component to g:
    //       u <- (n.u)n + g
    //
    // for TZ flow g = ue - (n.ue)n
    getVelocityComponents( n1,n2,n3,m1,m2,m3, u,bcParameters,"tangentialComponent",uC,fC  );

    if( numberOfDimensions==2 )
    {
      uDotN(I1,I2,I3)=uA(I1,I2,I3,n1)*normal(I1,I2,I3,0)+uA(I1,I2,I3,n2)*normal(I1,I2,I3,1);
    }
    else if( numberOfDimensions==3 )
    {
      uDotN(I1,I2,I3)=uA(I1,I2,I3,n1)*normal(I1,I2,I3,0)
	+uA(I1,I2,I3,n2)*normal(I1,I2,I3,1) 
	+uA(I1,I2,I3,n3)*normal(I1,I2,I3,2);
    }
    else
      uDotN(I1,I2,I3)=uA(I1,I2,I3,n1)*(2*side-1);  // outward normal in 1D
    if( twilightZoneFlow ) 
    { 
      if( numberOfDimensions==2 )
	uDotN(I1,I2,I3)-=(*e)(c,I1,I2,I3,m1,t)*normal(I1,I2,I3,0)
	  +(*e)(c,I1,I2,I3,m2,t)*normal(I1,I2,I3,1);
      else if( numberOfDimensions==3 ) 
	uDotN(I1,I2,I3)-=(*e)(c,I1,I2,I3,m1,t)*normal(I1,I2,I3,0)
	  +(*e)(c,I1,I2,I3,m2,t)*normal(I1,I2,I3,1) 
	  +(*e)(c,I1,I2,I3,m3,t)*normal(I1,I2,I3,2);
      else
	uDotN(I1,I2,I3)-=(*e)(c,I1,I2,I3,m1,t)*(2*side-1);  // outward normal in 1D

      if( numberOfDimensions>1 )
      {
	WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*normal(I1,I2,I3,0)+(*e)(c,I1,I2,I3,m1,t); )
	WHERE_MASK( uA(I1,I2,I3,n2)=uDotN(I1,I2,I3)*normal(I1,I2,I3,1)+(*e)(c,I1,I2,I3,m2,t); )
	if( numberOfDimensions==3 )
	  WHERE_MASK( uA(I1,I2,I3,n3)=uDotN(I1,I2,I3)*normal(I1,I2,I3,2)+(*e)(c,I1,I2,I3,m3,t); )
      }
      else
	WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*(2*side-1)+(*e)(c,I1,I2,I3,m1,t); )
    }
    else if( bcOption==scalarForcing )
    {
      if( scalarData == 0. )
      {
	if( numberOfDimensions>1 )
	{
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*normal(I1,I2,I3,0); )
	  WHERE_MASK( uA(I1,I2,I3,n2)=uDotN(I1,I2,I3)*normal(I1,I2,I3,1); )
	  if( numberOfDimensions==3 )
	    WHERE_MASK( uA(I1,I2,I3,n3)=uDotN(I1,I2,I3)*normal(I1,I2,I3,2); )
	}
	else
	  uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*(2*side-1);
      }
      else
      {
	if( numberOfDimensions>1 )
	{
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*normal(I1,I2,I3,0)+scalarData; )
	  WHERE_MASK( uA(I1,I2,I3,n2)=uDotN(I1,I2,I3)*normal(I1,I2,I3,1)+scalarData; )
	  if( numberOfDimensions==3 )
	    WHERE_MASK( uA(I1,I2,I3,n3)=uDotN(I1,I2,I3)*normal(I1,I2,I3,2)+scalarData; )
	}
	else
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*(2*side-1)+scalarData; )
      }
    }
    else if( bcOption==arrayForcing )
    {
      if( side<=arrayData.getBound(1) && axis<=arrayData.getBound(2) && grid<=arrayData.getBound(3) )
      {
	if( numberOfDimensions>1 )
	{
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*normal(I1,I2,I3,0)+arrayData(0,side,axis,grid); )
	  WHERE_MASK( uA(I1,I2,I3,n2)=uDotN(I1,I2,I3)*normal(I1,I2,I3,1)+arrayData(1,side,axis,grid); )
	  if( numberOfDimensions==3 )
	    WHERE_MASK( uA(I1,I2,I3,n3)=uDotN(I1,I2,I3)*normal(I1,I2,I3,2)+arrayData(2,side,axis,grid); )
	}
	else
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*(2*side-1)+arrayData(0,side,axis,grid); )
      }
      else
      {
	if( numberOfDimensions>1 )
	{
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*normal(I1,I2,I3,0)+arrayData(0); )
	  WHERE_MASK( uA(I1,I2,I3,n2)=uDotN(I1,I2,I3)*normal(I1,I2,I3,1)+arrayData(1); )
	  if( numberOfDimensions==3 )
	    WHERE_MASK( uA(I1,I2,I3,n3)=uDotN(I1,I2,I3)*normal(I1,I2,I3,2)+arrayData(2); )
	}
	else
	  WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*(2*side-1)+arrayData(0); )
      }
    }
    else if( bcOption==gridFunctionForcing )
    {  // use user supplied variable values
      if( numberOfDimensions>1 )
      {
	WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*normal(I1,I2,I3,0)+gfData(I1,I2,I3,m1); )
	WHERE_MASK( uA(I1,I2,I3,n2)=uDotN(I1,I2,I3)*normal(I1,I2,I3,1)+gfData(I1,I2,I3,m2); )
	if( numberOfDimensions==3 )
	  WHERE_MASK( uA(I1,I2,I3,n3)=uDotN(I1,I2,I3)*normal(I1,I2,I3,2)+gfData(I1,I2,I3,m3); )
      }
      else
	WHERE_MASK( uA(I1,I2,I3,n1)=uDotN(I1,I2,I3)*(2*side-1)+gfData(I1,I2,I3,m1); )

    }
    else
      {throw "Invalid value for bcOption! (tangentialComponent)";}

    break;
    
  }
}
